/*
	Copyright  1995-2001, The AROS Development Team. All rights reserved.
	$Id: adddatatypes.c,v 1.1.1.1 2005/03/15 16:01:02 laire Exp $
 
	Desc:
	Lang: English
*/

#define USE_SOFTLINKCHECK		1

//#include <aros/macros.h>
//#include <aros/bigendianio.h>
#include <exec/types.h>
#include <exec/memory.h>
#include <exec/execbase.h>
#include <dos/dos.h>
#include <dos/dosasl.h>
#include <dos/dosextens.h>
#include <libraries/iffparse.h>
#include <utility/name.h>
#include <utility/hooks.h>
#include <workbench/startup.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/utility.h>
#include <proto/iffparse.h>
#include <proto/alib.h>
//#include <proto/arossupport.h>

#include <string.h>
#include <stdio.h>

#include "../../libs/datatypes/datatypes_intern.h"
#undef DOSBase
#undef UtilityBase
#undef IFFParseBase
#undef LocaleBase
#undef DTList

#define	DEBUG(x)	;
#ifdef __MORPHOS__
#define	AROS_BE2WORD(x)	x
#endif
/******************************** STRUCTURES *********************************/

/* same as datatypes/datatypes.h/struct DataTypeHeader, but always big endian
and 32 bit pointers (which in the file are actually offsets) */

struct FileDataTypeHeader
{
	ULONG	 fdth_NameOffset;     /* Name of the data type */
	ULONG	 fdth_BaseNameOffset; /* Base name of the data type */
	ULONG	 fdth_PatternOffset;  /* File name match pattern */
	ULONG	 fdth_MaskOffset;     /* Comparision mask (binary) */
	ULONG	 fdth_GroupID;        /* DataType Group */
	ULONG	 fdth_ID;             /* DataType ID (same as IFF FORM type) */
	WORD	 fdth_MaskLen;        /* Length of the comparision mask */
	WORD	 fdth_Pad;            /* Unused at present (must be 0) */
	UWORD	 fdth_Flags;          /* Flags -- see below */
	UWORD	 fdth_Priority;
};

#define O(x)        offsetof(struct FileDataTypeHeader,x)

#if 0
static const IPTR FileDataTypeHeaderDesc[] =
	{
		sizeof(struct FileDataTypeHeader),
		SDM_ULONG(O(fdth_NameOffset)),
		SDM_ULONG(O(fdth_BaseNameOffset)),
		SDM_ULONG(O(fdth_PatternOffset)),
		SDM_ULONG(O(fdth_MaskOffset)),
		SDM_ULONG(O(fdth_GroupID)),
		SDM_ULONG(O(fdth_ID)),
		SDM_WORD(O(fdth_MaskLen)),
		SDM_WORD(O(fdth_Pad)),
		SDM_UWORD(O(fdth_Flags)),
		SDM_UWORD(O(fdth_Priority)),
		SDM_END
	};
#endif

/******************************** PROTOTYPES *********************************/

struct StackVars;		/* forward declaration */

BOOL DateScan(struct StackVars *sv, STRPTR path);
void ScanDirectory(struct StackVars *sv, STRPTR pattern);
struct DataTypesList *CreateDTList(struct StackVars *sv);
struct CompoundDatatype *CreateBasicType(struct StackVars *sv,
                                         struct List *list,
                                         struct List *globallist, STRPTR name,
                                         UWORD Flags, ULONG ID, ULONG GroupID);
void LoadDatatype(struct StackVars *sv, STRPTR name);
struct CompoundDatatype *CreateDatatype(struct StackVars *sv,
                                        struct IFFHandle *iff);
struct CompoundDatatype *AddDatatype(struct StackVars *sv,
                                     struct CompoundDatatype *cdt);
void DeleteDatatype(struct StackVars *sv, struct CompoundDatatype *cdt);
void AlphaInsert(struct StackVars *sv, struct List *list, struct Node *node);
void PrioInsert(struct StackVars *sv, struct List *list,
                struct CompoundDatatype *cdt);
struct Node *__FindNameNoCase(struct StackVars *sv, struct List *list,
                              STRPTR name);

LONG ReadFunc(struct StackVars *sv, UBYTE *buffer, ULONG length);
UBYTE *AllocFunc(ULONG size, ULONG flags);
void FreeFunc(UBYTE *memory, ULONG size);
struct NamedObject *allocnamedobject(struct StackVars *sv, STRPTR name,
                                     Tag FirstTag, ...);


/********************************* CONSTANTS *********************************/

#include "adddatatypes_version.h"

static const char version[] = VERSTAG;

#define EXCL_LEN 18
static const
UBYTE ExcludePattern[] = "#?.(info|backdrop)";

static const
UBYTE Template[] = "FILES/M,QUIET/S,REFRESH/S,LIST/S";

struct ArgArray
{
	UBYTE **aa_Files;
	IPTR    aa_Quiet;
	IPTR    aa_Refresh;
	IPTR    aa_List;
};

#define  ID_DTCD  MAKE_ID('D','T','C','D')
#define  ID_DTTL  MAKE_ID('D','T','T','L')

#define  NUM_PROP  2

static const
LONG PropArray[2*NUM_PROP]=
	{
		ID_DTYP, ID_DTHD,
		ID_DTYP, ID_DTCD
	};

#define  NUM_COLL  1

static const
LONG CollArray[2*NUM_COLL]=
	{
		ID_DTYP, ID_DTTL
	};


#ifdef __MORPHOS__

extern struct EmulLibEntry GATE_ReadFunc;
extern struct EmulLibEntry GATE_AllocFunc;
extern struct EmulLibEntry GATE_FreeFunc;

LONG_FUNC FunctionArray[]=
	{				/* Note! */
		(LONG_FUNC)&GATE_ReadFunc,
		(LONG_FUNC)&GATE_AllocFunc,
		(LONG_FUNC)&GATE_FreeFunc
	};
#else
LONG_FUNC FunctionArray[]=
{				/* Note! */
	(LONG_FUNC)&ReadFunc,
	(LONG_FUNC)&AllocFunc,
	(LONG_FUNC)&FreeFunc
};
#endif


struct StackVars
{
	struct DataTypesList *DTList;
	UBYTE                 ExclPat[2*EXCL_LEN+2+1];
	struct ArgArray  AA;
	UBYTE           *HookBuffer;
	ULONG            HookBufSize;
	ULONG            HookPosition;
};

#undef  SysBase
#define DTList            sv->DTList
#define ExclPat           sv->ExclPat
#define AA                sv->AA
#define HookBuffer        sv->HookBuffer
#define HookBufSize       sv->HookBufSize
#define HookPosition      sv->HookPosition


/****** AddDatatypes/main *****************************************************
*
*   NAME
*        main - well... main
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/


struct ExecBase	*SysBase;
struct DosLibrary	*DOSBase;
struct Library	*UtilityBase;
struct Library	*LocaleBase;
struct Library	*IFFParseBase;
struct Library	*DataTypesBase;


int main(void)
{
	struct StackVars  vars;
	struct StackVars *sv;
	int    result = RETURN_FAIL;
	struct WBStartup *WBenchMsg = NULL;

	DEBUG(dprintf(__FUNCTION__ ":\n"));

	SysBase=*((struct ExecBase**) 4);
	if ((DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37)))
	{
		if ((UtilityBase=OpenLibrary("utility.library",37)))
		{
			if ((IFFParseBase=OpenLibrary("iffparse.library",37)))
			{
				if ((LocaleBase=OpenLibrary("locale.library",37)))
				{
					if ((DataTypesBase=OpenLibrary("datatypes.library",44)))
					{
						memset(&vars, 0, sizeof(struct StackVars));
						sv = &vars;

						if((DTList = CreateDTList(sv)))
						{
							struct Process *MyProcess;

							DEBUG(dprintf(__FUNCTION__ ": DTList 0x%lx\n",
										  DTList));

							MyProcess = (struct Process*) FindTask(NULL);

							ParsePatternNoCase(ExcludePattern, ExclPat, sizeof(ExclPat));

							ObtainSemaphore(&DTList->dtl_Lock);

							if (MyProcess->pr_CLI == NULL)
							{
								UWORD num;
								struct WBArg *wa;

								DEBUG(dprintf(__FUNCTION__ ": WB\n"));

								WaitPort(&MyProcess->pr_MsgPort);

								WBenchMsg = (struct WBStartup *) GetMsg(&MyProcess->pr_MsgPort);
								if (WBenchMsg)
								{
									BPTR	origdir;

									origdir = CurrentDir(DupLock(WBenchMsg->sm_ArgList->wa_Lock));

									for (num = 1, wa = &WBenchMsg->sm_ArgList[1];
									     num < WBenchMsg->sm_NumArgs;
									     num++, wa++)
									{
										BPTR olddir = CurrentDir(wa->wa_Lock);
										LoadDatatype(sv, wa->wa_Name);
										(void) CurrentDir(olddir);
									}
									
									UnLock(CurrentDir(origdir));
								}
								result = RETURN_OK;
							}
							else
							{
								struct RDArgs *RDArgs;

								DEBUG(dprintf(__FUNCTION__ ": CLI\n"));

								if(!(RDArgs = ReadArgs(Template, (LONG*)&AA, NULL)))
								{
									PrintFault(IoErr(), NULL);
								}
								else
								{
									if (AA.aa_Refresh)
									{
										APTR OldWinPtr;
										BPTR moslock, syslock;

										DEBUG(dprintf(__FUNCTION__ ": Refresh\n"));

										OldWinPtr = MyProcess->pr_WindowPtr;
										MyProcess->pr_WindowPtr = (APTR) -1;

										syslock = Lock("DEVS:DataTypes/", ACCESS_READ);
										moslock = Lock("MOSSYS:Devs/DataTypes/", ACCESS_READ);
										if (syslock && moslock && SameLock(syslock, moslock) == LOCK_SAME)
										{
											UnLock(syslock);
											syslock = NULL;
										}

										/* Scan DEVS:DataTypes first since ScanDirectory *replace* previous
										 * matching descriptors... This way MOSSYS:Devs/DataTypes has priority.
										 * - Piru
										 */

										if (syslock)
										{
											if (DateScan(sv, "DEVS:DataTypes"))
											{
												DEBUG(dprintf(__FUNCTION__ ": DateScan ok\n"));
												ScanDirectory(sv, "DEVS:DataTypes");
											}
											else
											{
												DEBUG(dprintf(__FUNCTION__ ": DateScan failed\n"));
											}

											UnLock(syslock);
										}

										if (moslock)
										{
											if (DateScan(sv, "MOSSYS:Devs/DataTypes"))
											{
												DEBUG(dprintf(__FUNCTION__ ": DateScan ok\n"));
												ScanDirectory(sv, "MOSSYS:Devs/DataTypes");
											}
											else
											{
												DEBUG(dprintf(__FUNCTION__ ": DateScan failed\n"));
											}

											UnLock(moslock);
										}

										MyProcess->pr_WindowPtr = OldWinPtr;
									}
									else
									{
										UBYTE **files = AA.aa_Files;

										if(files)
										{
											while(*files)
											{
												ScanDirectory(sv, *files);
												files++;
											}
										}
									}

									if(AA.aa_List)
									{
										struct DataTypesList *dtl = NULL;
										struct NamedObject   *no  = NULL;

										DEBUG(dprintf(__FUNCTION__ ":\n"));

										if((no = FindNamedObject(NULL, DATATYPESLIST, NULL)))
										{
											dtl = (struct DataTypesList*)no->no_Object;
											ReleaseNamedObject(no);

											if(dtl != NULL)
											{
												struct Node *node=dtl->dtl_SortedList.lh_Head;

												while(node->ln_Succ != NULL)
												{
														// sorted list points to DT.dtn_Node2 ....
													struct CompoundDatatype *cdt;
													struct DataTypeHeader *dth;
													STRPTR argarray[2];

													if(CheckSignal(SIGBREAKF_CTRL_C))
													{
														Flush(Output());
														PrintFault(ERROR_BREAK,0);
														break;
													}
													cdt=(struct CompoundDatatype *)(node-1);
													dth=cdt->DT.dtn_Header;

													argarray[0] = dth->dth_BaseName;
													argarray[1] = dth->dth_Name;

													VPrintf("%s, \"%s\"\n",argarray);
													node = node->ln_Succ;
												}
											}
										}

									}

									DEBUG(dprintf(__FUNCTION__ ": free args\n"));
									result = RETURN_OK;
									FreeArgs(RDArgs);
								}
							}

							ReleaseSemaphore(&DTList->dtl_Lock);
						}
						CloseLibrary((struct Library *)DataTypesBase);
					}
					CloseLibrary((struct Library *)LocaleBase);
				}
				CloseLibrary((struct Library *)IFFParseBase);
			}
			CloseLibrary((struct Library *)UtilityBase);
		}
		CloseLibrary((struct Library *)DOSBase);
	}
	DEBUG(dprintf(__FUNCTION__ ": result 0x%lx\n",result));

	if (WBenchMsg)
	{
		Forbid();
		ReplyMsg((struct Message *) WBenchMsg);
	}

	return result;
}



/****** AddDatatypes/DateScan *************************************************
*
*   NAME
*        DateScan - See if datatypes descriptors need updating
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

BOOL DateScan(struct StackVars *sv, STRPTR path)
{
	BOOL   result = TRUE;
	BPTR   lock;
	struct FileInfoBlock *fib;

	if((lock = Lock(path, ACCESS_READ)))
	{
		if((fib = AllocDosObject(DOS_FIB, NULL)))
		{

			if(Examine(lock, fib))
			{
				struct DateStamp DirStamp;

				DirStamp = fib->fib_Date;

				if(ExNext(lock,fib)) // make sure there are files in this dir before updating datestamp to avoid problems
						     // with empty devs:datatypes dir and same mossys:devs/datatypes date
				{
					if(!CompareDates(&DirStamp, &DTList->dtl_DateStamp))
					{
						result = FALSE;
					}
					else
					{
						DTList->dtl_DateStamp = DirStamp;
					}
				}
			}

			FreeDosObject(DOS_FIB,fib);
		}

		UnLock(lock);
	}

	return result;
}



/****** AddDatatypes/ScanDirectory ********************************************
*
*   NAME
*        ScanDirectory - Scan a directory recursively for DT descriptors
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

void ScanDirectory(struct StackVars *sv, STRPTR pattern)
{
	struct AnchorPath *AnchorPath;
	LONG               RetVal;
	BPTR               OldDir;
	BOOL               DirEntered = FALSE;

	DEBUG(dprintf(__FUNCTION__ ": pattern \"%s\"\n", pattern));

	if((AnchorPath = (struct AnchorPath *)AllocVec(sizeof(struct AnchorPath),
					 MEMF_CLEAR)))
	{
		AnchorPath->ap_BreakBits = SIGBREAKF_CTRL_C;

		RetVal = MatchFirst(pattern, AnchorPath);

		while(!RetVal)
		{
			if(CheckSignal(SIGBREAKF_CTRL_C))
			{
				if(!AA.aa_Quiet)
				{
					PrintFault(ERROR_BREAK, NULL);
				}
				break;
			}

#if USE_SOFTLINKCHECK
			if (SoftlinkDODIR(AnchorPath, TRUE, AA.aa_Quiet, DOSBase))
#else /* USE_SOFTLINKCHECK */
			if(AnchorPath->ap_Info.fib_DirEntryType > 0L)
#endif /* USE_SOFTLINKCHECK */

			{
				DEBUG(dprintf(__FUNCTION__ ": dir \"%s\"\n", AnchorPath->ap_Info.fib_FileName));

				/* 27-Mar-2003 bugfix: AddDatatypes must NOT BE RECURSIVE!
				 * For patterns like "foo:bar/" the first dir must be entered. - Piru
				 */
				if (!DirEntered)
				{
					DirEntered = TRUE;

					if(!(AnchorPath->ap_Flags & APF_DIDDIR))
					{
						AnchorPath->ap_Flags |= APF_DODIR;
					}

					AnchorPath->ap_Flags &= ~APF_DIDDIR;
				}
			}
			else
			{
				DEBUG(dprintf(__FUNCTION__ ": file \"%s\"\n", AnchorPath->ap_Info.fib_FileName));

				if(!MatchPatternNoCase(ExclPat,
									   AnchorPath->ap_Info.fib_FileName))
				{
					OldDir = CurrentDir(AnchorPath->ap_Current->an_Lock);

					LoadDatatype(sv, AnchorPath->ap_Info.fib_FileName);

					CurrentDir(OldDir);
				}
			}
			RetVal = MatchNext(AnchorPath);
		}

		if(RetVal != ERROR_NO_MORE_ENTRIES)
		{
			if(!AA.aa_Quiet)
			{
				PrintFault(RetVal, NULL);
			}
		}

		DEBUG(dprintf(__FUNCTION__ ": MatchEnd\n"));

		MatchEnd(AnchorPath);

		FreeVec((APTR)AnchorPath);
	}
	DEBUG(dprintf(__FUNCTION__ ": done\n"));
}


/****** AddDatatypes/CreateDTList *********************************************
*
*   NAME
*        CreateDTList - Create and initialize the DataTypesList
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

struct DataTypesList *CreateDTList(struct StackVars *sv)
{
	struct DataTypesList *dtl = NULL;
	struct NamedObject   *no  = NULL;

	DEBUG(dprintf(__FUNCTION__ ":\n"));

	if((no = FindNamedObject(NULL, DATATYPESLIST, NULL)))
	{
		DEBUG(dprintf(__FUNCTION__ ": found NamedObject 0x%lx\n",no));
		dtl = (struct DataTypesList*)no->no_Object;
		DEBUG(dprintf(__FUNCTION__ ": dtl 0x%lx\n",dtl));
	}

#if 0
	if(!dtl)
	{
		if((no = AllocNamedObject(DATATYPESLIST,
								  ANO_NameSpace, TRUE,
								  ANO_UserSpace, sizeof(struct DataTypesList),
								  ANO_Flags    , NSF_NODUPS | NSF_CASE,
								  TAG_DONE)))
		{
			DEBUG(dprintf(__FUNCTION__ ": created NamedObject 0x%lx\n",no));
			if (!(dtl = (struct DataTypesList*)no->no_Object))
			{
				FreeNamedObject(no);
				no = NULL;
			}
		}

		if(dtl)
		{

			DEBUG(dprintf(__FUNCTION__ ": dtl 0x%lx\n",dtl));

			InitSemaphore(&dtl->dtl_Lock);

			NewList(&dtl->dtl_SortedList);
			NewList(&dtl->dtl_BinaryList);
			NewList(&dtl->dtl_ASCIIList);
			NewList(&dtl->dtl_IFFList);
			NewList(&dtl->dtl_MiscList);

			if(!AddNamedObject(NULL, no))
			{
				FreeNamedObject(no);
				no = NULL;
				dtl = NULL;
			}
		}
	}
#endif
	if(dtl)
	{
		if(!__FindNameNoCase(sv, &dtl->dtl_BinaryList, "binary"))
		{
			CreateBasicType(sv, &dtl->dtl_BinaryList, &dtl->dtl_SortedList,
							"binary", DTF_BINARY,
							MAKE_ID('b','i','n','a'),
							MAKE_ID('s','y','s','t'));
		}

		if(!__FindNameNoCase(sv, &dtl->dtl_ASCIIList, "ascii"))
		{
			CreateBasicType(sv, &dtl->dtl_ASCIIList,  &dtl->dtl_SortedList,
							"ascii", DTF_ASCII,
							MAKE_ID('a','s','c','i'),
							MAKE_ID('t','e','x','t'));
		}

		if(!__FindNameNoCase(sv, &dtl->dtl_IFFList, "iff"))
		{
			CreateBasicType(sv, &dtl->dtl_IFFList, &dtl->dtl_SortedList,
							"iff", DTF_IFF,
							MAKE_ID('i','f','f', 0 ),
							MAKE_ID('s','y','s','t'));
		}

		if(!__FindNameNoCase(sv, &dtl->dtl_MiscList, "directory"))
		{
			CreateBasicType(sv, &dtl->dtl_MiscList, &dtl->dtl_SortedList,
							"directory", DTF_MISC,
							MAKE_ID('d','i','r','e'),
							MAKE_ID('s','y','s','t'));
		}
	}

	if(no)
	{
		ReleaseNamedObject(no);
	}


	return dtl;
}



/****** AddDatatypes/CreateBasicType ******************************************
*
*   NAME
*        CreateBasicType - Initialize one of the basic types
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

struct CompoundDatatype *CreateBasicType(struct StackVars *sv,
                                         struct List *list,
                                         struct List *globallist, STRPTR name,
                                         UWORD Flags, ULONG ID, ULONG GroupID)
{
	struct CompoundDatatype *cdt;
	ULONG AllocLen = sizeof(struct CompoundDatatype) + strlen(name) + 1;

	DEBUG(dprintf(__FUNCTION__ "\n"));

	if((cdt = AllocVec(AllocLen, MEMF_PUBLIC | MEMF_CLEAR)))
	{
		cdt->DT.dtn_Header = &cdt->DTH;

		strcpy((UBYTE*)(cdt + 1), name);

		cdt->DTH.dth_Name=
			cdt->DTH.dth_BaseName=
				cdt->DT.dtn_Node1.ln_Name=
					cdt->DT.dtn_Node2.ln_Name=(UBYTE*)(cdt + 1);

		cdt->DTH.dth_GroupID = GroupID;
		cdt->DTH.dth_ID = ID;

		cdt->DTH.dth_Flags = Flags;

		NewList(&cdt->DT.dtn_ToolList);

		cdt->DT.dtn_Length = AllocLen;

		AddTail(list, &cdt->DT.dtn_Node1);

		AlphaInsert(sv, globallist, &cdt->DT.dtn_Node2);
	}

	return cdt;
}



/****** AddDatatypes/LoadDatatype *********************************************
*
*   NAME
*        LoadDatatype - Load and install a single datatype descriptor
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

void LoadDatatype(struct StackVars *sv, STRPTR name)
{
	struct IFFHandle *iff;

	DEBUG(dprintf(__FUNCTION__ ": <%s>\n",name));

	if((iff = AllocIFF()))
	{
		DEBUG(dprintf(__FUNCTION__ "<%s>\n",name));
		if((iff->iff_Stream = (IPTR)Open(name, MODE_OLDFILE))) /* Why IPTR? */
		{
			InitIFFasDOS(iff);

			if (!OpenIFF(iff, IFFF_READ))
			{
				if(!PropChunks(iff, PropArray, NUM_PROP))
				{
					if(!CollectionChunks(iff, CollArray, NUM_COLL))
					{
						if(!StopOnExit(iff, ID_DTYP, ID_FORM))
						{
							LONG error;

							while((error = ParseIFF(iff, IFFPARSE_SCAN)) == IFFERR_EOC)
							{
								CreateDatatype(sv, iff);
#warning The while ParseIFF loop here crashes the 2nd time inside the loop, therefore the break below as temp fix
								break;
							}
						}
						else
						{
							DEBUG(dprintf(__FUNCTION__ ": StopOnExit failed\n"));
						}
					}
					else
					{
						DEBUG(dprintf(__FUNCTION__ ": CollectionChunks failed\n"));
					}
				}
				else
				{
					DEBUG(dprintf(__FUNCTION__ ": PropChunks failed\n"));
				}
				CloseIFF(iff);
			}
			else
			{
				DEBUG(dprintf(__FUNCTION__ ": OpenIFF failed\n"));
			}
			Close((BPTR)iff->iff_Stream);
		}
		else
		{
			DEBUG(dprintf(__FUNCTION__ ": Open failed\n"));
		}
		FreeIFF(iff);
	}
	else
	{
		SetIoErr(ERROR_NO_FREE_STORE);
	}
}

#if 0
/****** AddDatatypes/MemStreamHook *******************************************
*
*   NAME
*        MemStreamHook - needed by ReadStruct
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

LONG MemStreamHook(struct Hook * hook, UBYTE **memptr, Msg msg)
{
	LONG rc;

	switch (msg->MethodID)
	{
	case BEIO_READ:
		rc = **memptr;
		(*memptr)++;
		break;

	default:
		rc = EOF;
		break;
	}

	return rc;
}
#endif

/****** AddDatatypes/CreateDatatype *******************************************
*
*   NAME
*        CreateDatatype - create a datatype from IFF chunks
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

struct CompoundDatatype *CreateDatatype(struct StackVars *sv,
                                        struct IFFHandle *iff)
{
	struct CompoundDatatype *cdt = NULL;
	struct StoredProperty *prop;
	ULONG  AllocLen;
	UBYTE *func;
	LONG   DefaultStack = AROS_STACKSIZE, i;
	BPTR   SegList;

	DEBUG(dprintf(__FUNCTION__ "\n"));

	if((prop = FindProp(iff, ID_DTYP, ID_DTHD)))
	{
		AllocLen = sizeof(struct CompoundDatatype) -
				   32 + /* was sizeof(struct DataTypeHeader), but we must use struct size as it would be on Amiga */
				   prop->sp_Size;

		if(!(cdt = AllocVec(AllocLen, MEMF_PUBLIC | MEMF_CLEAR)))
		{
			SetIoErr(ERROR_NO_FREE_STORE);
		}
		else
		{
			struct FileDataTypeHeader *fdh;
			IPTR extraoffset = sizeof(struct DataTypeHeader) - 32;

			fdh = (struct FileDataTypeHeader *) prop->sp_Data;


			cdt->DT.dtn_Header= &cdt->DTH;

			cdt->DTH.dth_Name = (STRPTR)(fdh->fdth_NameOffset + extraoffset + (IPTR)&cdt->DTH);
			cdt->DTH.dth_BaseName = (STRPTR)(fdh->fdth_BaseNameOffset + extraoffset + (IPTR)&cdt->DTH);
			cdt->DTH.dth_Pattern = (STRPTR)(fdh->fdth_PatternOffset + extraoffset + (IPTR)&cdt->DTH);
			cdt->DTH.dth_Mask = (WORD *)(fdh->fdth_MaskOffset + extraoffset + (IPTR)&cdt->DTH);
			cdt->DTH.dth_GroupID = fdh->fdth_GroupID;
			cdt->DTH.dth_ID = fdh->fdth_ID;
			cdt->DTH.dth_MaskLen = fdh->fdth_MaskLen;
			cdt->DTH.dth_Pad = fdh->fdth_Pad;
			cdt->DTH.dth_Flags = fdh->fdth_Flags;
			cdt->DTH.dth_Priority = fdh->fdth_Priority;

			CopyMem(prop->sp_Data + 32, cdt + 1, prop->sp_Size - 32);

			for(i = 0; i < cdt->DTH.dth_MaskLen; i++)
			{
				cdt->DTH.dth_Mask[i] = AROS_BE2WORD(cdt->DTH.dth_Mask[i]);
				DEBUG(dprintf("mask[%d] = %04x (%c %c)\n", i,
							  cdt->DTH.dth_Mask[i],
							  cdt->DTH.dth_Mask[i] & 255,
							  (cdt->DTH.dth_Mask[i] >> 8) & 255));
			}

			DEBUG(dprintf("groupid  = %c%c%c%c\n", cdt->DTH.dth_GroupID >> 24,
						  cdt->DTH.dth_GroupID >> 16,
						  cdt->DTH.dth_GroupID >> 8,
						  cdt->DTH.dth_GroupID));
			DEBUG(dprintf("id       = %c%c%c%c\n", cdt->DTH.dth_ID >> 24,
						  cdt->DTH.dth_ID >> 16,
						  cdt->DTH.dth_ID >> 8,
						  cdt->DTH.dth_ID));
			DEBUG(dprintf("flags    = %x\n",	 cdt->DTH.dth_Flags));
			DEBUG(dprintf("pri      = %d\n",	 cdt->DTH.dth_Priority));
			DEBUG(dprintf("name     = %s\n",	 cdt->DTH.dth_Name));
			DEBUG(dprintf("basename = %s\n",	 cdt->DTH.dth_BaseName));
			DEBUG(dprintf("pattern  = %s\n",	 cdt->DTH.dth_Pattern));
			DEBUG(dprintf("masklen  = %d\n",	 cdt->DTH.dth_MaskLen));

			NewList(&cdt->DT.dtn_ToolList);

			cdt->DT.dtn_Length = AllocLen;

#if 1
			if((prop = FindProp(iff, ID_DTYP, ID_DTCD)))
			{
				DEBUG(dprintf("prop 0x%lx size %ld\n",prop,prop->sp_Size));
				if((func = AllocVec(prop->sp_Size, MEMF_PUBLIC | MEMF_CLEAR)))
				{
					DEBUG(dprintf("func buffer 0x%lx\n",func));
					cdt->DTCDChunk = func;
					cdt->DTCDSize = prop->sp_Size;

					CopyMem(prop->sp_Data,func,prop->sp_Size);

					HookBuffer = cdt->DTCDChunk;
					HookBufSize = cdt->DTCDSize;
					HookPosition = 0;

					if((SegList = InternalLoadSeg((BPTR)sv, (BPTR) NULL,
												  (ULONG *)FunctionArray,
												  &DefaultStack)))
					{
						DEBUG(dprintf("seglist 0x%lx\n",SegList));
						cdt->SegList = SegList;
						cdt->Function = (APTR)((((ULONG)SegList) << 2) + 4);
						DEBUG(dprintf("Function 0x%lx\n",cdt->Function));
					}
					else
					{
						DEBUG(dprintf("internalloadseg failed\n"));
						FreeVec(func);
						FreeVec(cdt);
						cdt = NULL;
					}

				} /* if((func = AllocVec(prop->sp_Size, MEMF_PUBLIC | MEMF_CLEAR))) */

			} /* if((prop = FindProp(iff, ID_DTYP, ID_DTCD))) */
#endif
			if (cdt)
			{
				DEBUG(dprintf(__FUNCTION__ ": add datatype\n"));
				cdt = AddDatatype(sv, cdt);
			}
			else
			{
				DEBUG(dprintf(__FUNCTION__ ": skipped datatype\n"));
			}
		} /* cdt AllocVec okay */

	} /* if((prop = FindProp(iff, ID_DTYP, ID_DTHD))) */

	return cdt;
}



/****** AddDatatypes/AddDatatype **********************************************
*
*   NAME
*        AddDatatype - add a datatype to the system
*
*   SYNOPSIS
*
*   FUNCTION
*
*        This subroutine tries to add a datatype to the system datatypes
*        list. If the datatype already exists, it will be replaced or
*        updated. In case of an error, the CompoundDatatype will be deleted
*        and a NULL pointer is returned.
*
*        The CompoundDatatype pointer you passed in will be invalid after
*        calling this function. Use the returned handle instead.
*        DO NOT USE THE OLD POINTER IN ANY MORE!
*
*   INPUTS
*
*   RETURNS
*
*        A pointer to a CompoundDatatype in the system datatypes list
*        or a NULL pointer for failure
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

struct CompoundDatatype *AddDatatype(struct StackVars *sv,
                                     struct CompoundDatatype *cdt)
{
	struct List *typelist;
	BOOL   Success = FALSE;
	ULONG  AllocSize;
	ULONG  result;
	struct CompoundDatatype *oldcdt;

	switch(cdt->DTH.dth_Flags & DTF_TYPE_MASK)
	{
	case DTF_BINARY:
		typelist= &DTList->dtl_BinaryList;
		break;
	case DTF_ASCII:
		typelist= &DTList->dtl_ASCIIList;
		break;
	case DTF_IFF:
		typelist= &DTList->dtl_IFFList;
		break;
	case DTF_MISC:
		typelist= &DTList->dtl_MiscList;
		break;
	default:
		typelist= NULL;
	}

	if(typelist)
	{
		cdt->DT.dtn_Node1.ln_Name = cdt->DT.dtn_Node2.ln_Name = cdt->DTH.dth_Name;

		Success = TRUE;

		AllocSize = strlen(cdt->DTH.dth_Pattern);

		if((!Stricmp(cdt->DTH.dth_Pattern, "#?")) ||
				(!AllocSize) )
		{
			cdt->FlagLong |= CFLGF_PATTERN_UNUSED;
		}
		else
		{
			cdt->FlagLong &= ~(CFLGF_PATTERN_UNUSED);

			AllocSize = 2 * AllocSize + 2;

			if(!(cdt->ParsePatMem = AllocVec(AllocSize,
											 MEMF_PUBLIC | MEMF_CLEAR)))
			{
				Success = FALSE;
			}
			else
			{
				cdt->ParsePatSize = AllocSize;

				result = ParsePatternNoCase(cdt->DTH.dth_Pattern,
											cdt->ParsePatMem, AllocSize);

				if(result == 1)
				{
					cdt->FlagLong |= CFLGF_IS_WILD;
				}
				else
				{
					FreeVec(cdt->ParsePatMem);
					cdt->ParsePatMem = NULL;
					cdt->ParsePatSize = 0;

					if(result == 0)
					{
						cdt->FlagLong &= ~(CFLGF_IS_WILD);
					}
					else
					{
						Success = FALSE;
					}
				}
			}
		}

		if (Success)
		{
			if((oldcdt = (struct CompoundDatatype*)__FindNameNoCase(sv,
						 typelist,
						 cdt->DT.dtn_Node1.ln_Name)))
			{
				if (oldcdt->OpenCount)
				{
					Success = FALSE;
				}
				else
				{
					if((Stricmp(oldcdt->DTH.dth_Name, cdt->DTH.dth_Name)) ||
							(Stricmp(oldcdt->DTH.dth_BaseName, cdt->DTH.dth_BaseName)) ||
							(Stricmp(oldcdt->DTH.dth_Pattern, cdt->DTH.dth_Pattern)) ||
							(oldcdt->DTH.dth_Flags != cdt->DTH.dth_Flags) ||
							(oldcdt->DTH.dth_Priority != cdt->DTH.dth_Priority) ||
							(oldcdt->DTH.dth_MaskLen != cdt->DTH.dth_MaskLen))
					{
						DeleteDatatype(sv, oldcdt);
						oldcdt = NULL;
					}
					else
					{
						oldcdt->DTH.dth_GroupID = cdt->DTH.dth_GroupID;
						oldcdt->DTH.dth_ID      = cdt->DTH.dth_ID;
						CopyMem(cdt->DTH.dth_Mask,cdt->DTH.dth_Mask,
								(ULONG)(sizeof(WORD)*cdt->DTH.dth_MaskLen));
					}
				}
			}

			if(Success)
			{
				if (oldcdt)
				{
					DeleteDatatype(sv, cdt);
					cdt = oldcdt;
				}
				else
				{
					if (cdt->DT.dtn_FunctionName)
					{
						LONG DefaultStack = 4096;
						BPTR file;
						ULONG AllocLen;
						BPTR SegList;

						DEBUG(dprintf("FunctionName <%s>\n",cdt->DT.dtn_FunctionName));

						if((file = Open(cdt->DT.dtn_FunctionName, MODE_OLDFILE)))
						{
							if(Seek(file, 0, OFFSET_END) >= 0)
							{
								if((AllocLen = Seek(file, 0,
													OFFSET_BEGINNING)) > 0)
								{
									DEBUG(dprintf("Alloclen %ld\n",AllocLen));
									if((cdt->DTCDChunk = AllocVec(AllocLen,
																  MEMF_PUBLIC | MEMF_CLEAR)))
									{
										DEBUG(dprintf("DTCDChunk 0x%lx\n",cdt->DTCDChunk));
										cdt->DTCDSize = AllocLen;

										if (Read(file, cdt->DTCDChunk, AllocLen) == AllocLen)
										{
											DEBUG(dprintf("read ok\n"));
											HookBuffer = cdt->DTCDChunk;
											HookBufSize = cdt->DTCDSize;
											HookPosition = 0;

											if((SegList = InternalLoadSeg((BPTR)sv, (BPTR)NULL, (ULONG *)FunctionArray, &DefaultStack)))
											{
												DEBUG(dprintf("seglist 0x%lx\n",SegList));
												cdt->SegList = SegList;
												cdt->Function = (APTR)((((ULONG)SegList)<<2)+4);
												DEBUG(dprintf("Function 0x%lx\n",cdt->Function));
											}
											else
											{
												DEBUG(dprintf("internalloadseg failed\n"));
											}
										}
										else
										{
											DEBUG(dprintf("read failed\n"));
											FreeVec(cdt->DTCDChunk);
											cdt->DTCDChunk = NULL;
											cdt->DTCDSize = 0;
										}
									}
								}
							}
						}
						cdt->DT.dtn_FunctionName=NULL;
					}

					if(cdt->DTH.dth_MaskLen > DTList->dtl_LongestMask)
					{
						DTList->dtl_LongestMask = cdt->DTH.dth_MaskLen;
					}

					PrioInsert(sv, typelist, cdt);

					AlphaInsert(sv, &DTList->dtl_SortedList, &cdt->DT.dtn_Node2);
				}
			}
		}
	}

	if(!Success)
	{
		DeleteDatatype(sv, cdt);
		cdt = NULL;
	}

	return cdt;
}



/****** AddDatatypes/DeleteDatatype *******************************************
*
*   NAME
*        DeleteDatatype - unlink and deallocate a CompoundDatatype structure
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

void DeleteDatatype(struct StackVars *sv, struct CompoundDatatype *cdt)
{
	if(cdt)
	{
		if(cdt->ParsePatMem)
		{
			FreeVec(cdt->ParsePatMem);
			cdt->ParsePatMem = NULL;
			cdt->ParsePatSize = 0;
		}

		if(cdt->DTCDChunk)
		{
			FreeVec(cdt->DTCDChunk);
			cdt->DTCDChunk = NULL;
			cdt->DTCDSize = 0;
		}

		if(cdt->SegList)
		{
			UnLoadSeg(cdt->SegList);
			cdt->SegList = NULL;
			cdt->Function = NULL;
		}

		if(cdt->DT.dtn_Node1.ln_Succ && cdt->DT.dtn_Node1.ln_Pred)
		{
			Remove(&cdt->DT.dtn_Node1);
			Remove(&cdt->DT.dtn_Node2);
			cdt->DT.dtn_Node1.ln_Succ = cdt->DT.dtn_Node1.ln_Pred =
											cdt->DT.dtn_Node2.ln_Succ = cdt->DT.dtn_Node2.ln_Pred = NULL;
		}

		FreeVec(cdt);
	}
}



/****** AddDatatypes/AlphaInsert **********************************************
*
*   NAME
*        AlphaInsert - enqueue a node alphabetically into a list
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

void AlphaInsert(struct StackVars *sv, struct List *list, struct Node *node)
{
	struct Node *cur,*prev=NULL;

	for(cur = list->lh_Head; cur->ln_Succ; prev = cur, cur = cur->ln_Succ)
	{
		if(Stricmp(cur->ln_Name, node->ln_Name) > 0)
			break;
	}

	Insert(list, node, prev);
}



/****** AddDatatypes/PrioInsert **********************************************
*
*   NAME
*        PrioInsert - enqueue a CompoundDatatype correctly in the type list
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

void PrioInsert(struct StackVars *sv, struct List *list,
                struct CompoundDatatype *cdt)
{
	struct CompoundDatatype *cur, *prev = NULL;
	WORD diff;

	for(cur = (struct CompoundDatatype*)list->lh_Head;
			cur->DT.dtn_Node1.ln_Succ;
			prev = cur, cur = (struct CompoundDatatype*)cur->DT.dtn_Node1.ln_Succ)
	{
		diff = (cdt->Function ? 1 : 0) - (cur->Function ? 1 : 0);

		if(diff > 0)
			break;

		if(!diff)
		{
			UWORD MinMask = (cdt->DTH.dth_MaskLen < cur->DTH.dth_MaskLen) ?
							cdt->DTH.dth_MaskLen : cur->DTH.dth_MaskLen;
			WORD *cdtmask = cdt->DTH.dth_Mask;
			WORD *curmask=cur->DTH.dth_Mask;

			while(!diff && MinMask--)
				diff= *(curmask++) - *(cdtmask++);

			if(diff > 0)
				break;

			if(!diff)
			{
				diff = cdt->DTH.dth_MaskLen - cur->DTH.dth_MaskLen;

				if(diff > 0)
					break;

				if(!diff)
				{
					diff = (((cdt->FlagLong & CFLGF_PATTERN_UNUSED) || cdt->DTH.dth_Pattern==NULL) ? 0 : 1) -
						   (((cur->FlagLong & CFLGF_PATTERN_UNUSED) || cur->DTH.dth_Pattern==NULL) ? 0 : 1);

					if(diff > 0)
						break;

					if(!diff)
					{
						diff = cdt->DTH.dth_Priority - cur->DTH.dth_Priority;

						if(diff > 0)
							break;
					}
				}
			}
		}
	}

	Insert(list, &cdt->DT.dtn_Node1, (struct Node *)prev);
}



/****** AddDatatypes/__FindNameNoCase *****************************************
*
*   NAME
*        __FindNameNoCase - find a node in a list (case insensitive)
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

struct Node *__FindNameNoCase(struct StackVars *sv, struct List *list,
                              STRPTR name)
{
	struct Node *node;
	struct Node *result = NULL;

	for(node = list->lh_Head; node->ln_Succ; node = node->ln_Succ)
	{
		if(!Stricmp(node->ln_Name, name))
		{
			result = node;
			break;
		}
	}

	return result;
}



/****** AddDatatypes/ReadFunc *************************************************
*
*   NAME
*        ReadFunc - data read hook for InternalLoadSeg
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

#ifdef __MORPHOS__

LONG ReadFuncEmul(void)
{
	struct StackVars *sv=(struct StackVars*) REG_D1;
	UBYTE *buffer=(UBYTE*) REG_D2;
	ULONG length=REG_D3;

	DEBUG(dprintf("ReadFuncEmul sv 0x%lx buffer 0x%lx length 0x%lx\n",sv,buffer,length));

	return(ReadFunc(sv,buffer,length));
}

struct EmulLibEntry GATE_ReadFunc=
	{
		TRAP_LIB,
		0,
		(void*) ReadFuncEmul
	};

#endif

LONG ReadFunc(struct StackVars *sv, UBYTE *buffer, ULONG length)
{
	LONG maxlen = HookBufSize-HookPosition;
	LONG actual = length > maxlen ? maxlen : length;

	CopyMem(HookBuffer+HookPosition, buffer, actual);

	HookPosition += actual;

	return actual;
}



/****** AddDatatypes/AllocFunc ************************************************
*
*   NAME
*        AllocFunc - memory allocation hook for InternalLoadSeg
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

#ifdef __MORPHOS__

UBYTE *AllocFuncEmul(void)
{
	ULONG size=REG_D0;
	ULONG flags=REG_D1;

	DEBUG(dprintf("AllocFuncEmul size 0x%lx flags 0x%lx\n",size,flags));

	return(AllocFunc(size,flags));
}

struct EmulLibEntry GATE_AllocFunc=
	{
		TRAP_LIB,
		0,
		(void*) AllocFuncEmul
	};


#endif

UBYTE *AllocFunc(ULONG size, ULONG flags)
{
	return(AllocMem(size, flags));
}



/****** AddDatatypes/FreeFunc *************************************************
*
*   NAME
*        FreeFunc - memory freeing hook for InternalLoadSeg
*
*   SYNOPSIS
*
*   FUNCTION
*
*   INPUTS
*
*   RETURNS
*
*   EXAMPLE
*
*   SEE ALSO
*
******************************************************************************
*
*/

#ifdef __MORPHOS__

void FreeFuncEmul(void)
{
	UBYTE *memory=(UBYTE*) REG_A1;
	ULONG size=REG_D0;

	DEBUG(dprintf("FreeFuncEmul memory 0x%lx size 0x%lx\n",memory,size));

	return(FreeFunc(memory,size));
}

struct EmulLibEntry GATE_FreeFunc=
	{
		TRAP_LIB,
		0,
		(void*) FreeFuncEmul
	};

#endif

void FreeFunc(UBYTE *memory, ULONG size)
{
	FreeMem(memory, size);
}


#if 0
/******************************* STUB ROUTINES ********************************/

struct NamedObject *allocnamedobject(struct StackVars *sv, STRPTR name,
												 Tag FirstTag, ...)
{
	return AllocNamedObjectA(name, (struct TagItem*)&FirstTag);
}
#endif
